@@ -13,18 +13,20 @@
 #    You should have received a copy of the GNU Lesser General Public License
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-__version__ = '0.8.2'
+__version__ = '0.9.0'
 
+import inspect
 import os
 import warnings
 
 from new import classobj
 
-__all__ = ['ArgvalidateException', 'DecoratorLengthException',
+__all__ = ['ArgvalidateException',
     'DecoratorNonKeyLengthException', 'DecoratorKeyUnspecifiedException',
     'DecoratorStackingException', 'ArgumentTypeException',
-    'ArgumentLengthException', 'func_args', 'method_args', 'return_value',
-    'one_of', 'type_any', 'raises_exceptions', 'warns_kwarg_as_arg']
+    'func_args', 'method_args', 'return_value',
+    'one_of', 'type_any', 'raises_exceptions', 'warns_kwarg_as_arg',
+    'accepts', 'returns']
 
 # Check for environment variables
 argvalidate_warn = 0
@@ -45,90 +47,16 @@ if 'ARGVALIDATE_WARN_KWARG_AS_ARG' in os.environ:
     except ValueError:
         pass
 
-class ArgvalidateErrorException(Exception):
+class ArgvalidateException(Exception):
     """
     Base argvalidate exception.
 
-    Used as base for all exceptions in exception mode.
-
-    .. note::
-
-       Do not use this exception directly. Always use ArgvalidateException
-       when checking for argvalidate errors.
+    Used as base for all exceptions.
 
     """
     pass
 
-class ArgvalidateErrorWarning(Warning):
-    """
-    Base argvalidate warning.
 
-    Used as base for all warnings in warning mode.
-
-    .. note::
-
-       Do not use this warning directly. Always use ArgvalidateException
-       when checking for argvalidate errors.
-
-    """
-    pass
-
-if argvalidate_warn:
-    ArgvalidateException = ArgvalidateErrorWarning
-else:
-    ArgvalidateException = ArgvalidateErrorWarning
-
-class ArgumentLengthException(ArgvalidateException):
-    """
-    Exception for invalid function-call argument count.
-
-    This exception provides the following attributes:
-
-    * func_name
-        Name of function that caused the exception to be raised
-        (str, read-only).
-
-    * expected_count
-        Number of arguments that were expected (int, read-only).
-
-    * passed_count
-        Number of arguments that were passed to the function (int, read-only).
-        
-    """
-    def __init__(self, func_name, expected_count, passed_count):
-        self.func_name = func_name
-        self.expected_count = expected_count
-        self.passed_count = passed_count
-
-        msg = '%s: wrong number of arguments passed to function (expected %d, got %d).' %\
-            (func_name, expected_count, passed_count)
-        ArgvalidateException.__init__(self, msg)
-
-class DecoratorLengthException(ArgvalidateException):
-    """
-    Exception for invalid decorator argument count.
-
-    This exception provides the following attributes:
-
-    * func_name
-        Name of function that caused the exception to be raised
-        (str, read-only).
-
-    * expected_count
-        Number of arguments that were expected (int, read-only).
-
-    * passed_count
-        Number of arguments that were passed to the function (int, read-only).
-
-    """
-    def __init__(self, func_name, expected_count, passed_count):
-        self.func_name = func_name
-        self.expected_count = expected_count
-        self.passed_count = passed_count
-        msg = '%s: wrong number of arguments specified in decorator ' %\
-            (func_name)
-        msg += '(expected %d, got %d).' % (expected_count, passed_count)
-        ArgvalidateException.__init__(self, msg)
 
 class DecoratorNonKeyLengthException(ArgvalidateException):
     """
@@ -255,7 +183,7 @@ class ReturnValueTypeException(ArgvalidateException):
             (func_name, expected_type, passed_type)
         ArgvalidateException.__init__(self, msg)
 
-class KWArgAsArgWarning(ArgvalidateErrorWarning):
+class KWArgAsArgWarning(ArgvalidateException):
     def __init__(self, func_name, arg_name):
         msg = '%s: argument %s is a keyword argument and was passed as a '\
             'non-keyword argument.' % (func_name, arg_name)
@@ -310,112 +238,182 @@ def __check_type(func_name, arg_name, expected_type, passed_value,\
         __raise(ArgumentTypeException(func_name, arg_name, expected_type,\
             passed_type), stacklevel=stacklevel)
 
-def __check_args(type_args, type_kwargs, start=0):
+def __check_args(type_args, type_kwargs, start=-1):
+    type_nonkey_argcount = len(type_args)
+    type_key_argcount = len(type_kwargs)
+
     def validate(f):
-        args_func = getattr(f, '_args_stacked_func', None)
-        if args_func:
+        accepts_func = getattr(f, 'argvalidate_accepts_stacked_func', None)
+        
+        if accepts_func:
+            if start == -1:
+                raise DecoratorStackingException(accepts_func.func_name,\
+                    'accepts')
             if start == 0:
-                raise DecoratorStackingException(args_func.func_name,\
-                     'func_args')
+                raise DecoratorStackingException(accepts_func.func_name,\
+                     'function_accepts')
             elif start == 1:
-                raise DecoratorStackingException(args_func.func_name,\
-                     'method_args')
+                raise DecoratorStackingException(accepts_func.func_name,\
+                     'method_accepts')
             else:
-                raise DecoratorStackingException(args_func.func_name,\
+                raise DecoratorStackingException(accepts_func.func_name,\
                      'unknown; start=%d' % (start))
 
-        func = getattr(f, '_return_value_stacked_func', f)
+        func = getattr(f, 'argvalidate_returns_stacked_func', f)
+        f_name = func.__name__
+        (func_args, func_varargs, func_varkw, func_defaults) =\
+             inspect.getargspec(func)
 
-        func_argcount = func.func_code.co_argcount - start
+        func_argcount = len(func_args)
+        is_method = True
+
+        # The original idea was to use inspect.ismethod here,
+        # but it seems as the decorator is called before the
+        # method is bound to a class, so this will always
+        # return False.
+        # The new method follows the original idea of checking
+        # tha name of the first argument passed.
+        # self and cls indicate methods, everything else indicates
+        # a function.
+        if start < 0 and func_argcount > 0 and func_args[0] in ['self', 'cls']:
+            func_argcount -= 1
+            func_args = func_args[1:]
+        elif start == 1:
+            func_argcount -=1
+            func_args = func_args[1:]
+        else:
+            is_method = False
+
+        if func_varargs:
+            func_args.remove(func_varargs)
+            
+        if func_varkw:
+            func_args.remove(func_varkw)
+
+        func_key_args = {}
         func_key_argcount = 0
-        f_name = func.func_name
-        
-        if func.func_defaults:
-            func_key_argcount = len(func.func_defaults)
 
-        func_nonkey_argcount = func_argcount - func_key_argcount
+        if func_defaults:
+            func_key_argcount = len(func_defaults)
+            tmp_key_args = zip(func_args[-func_key_argcount:], func_defaults)
+
+            for tmp_key_name, tmp_key_default in tmp_key_args:
+                func_key_args.update({tmp_key_name: tmp_key_default})
+
+            # Get rid of unused variables
+            del tmp_key_args
+            del tmp_key_name
+            del tmp_key_default
+
+        func_nonkey_args = []
+        if func_key_argcount < func_argcount:
+            func_nonkey_args = func_args[:func_argcount-func_key_argcount]
+        func_nonkey_argcount = len(func_nonkey_args)
+
+        # Static check #0:
+        #
+        # Checking the lengths of type_args vs. func_args and type_kwargs vs.
+        # func_key_args should be done here.
+        #
+        # This means that the check is only performed when the decorator
+        # is actually invoked, not every time the target function is called.
+        if func_nonkey_argcount != type_nonkey_argcount:
+            __raise(DecoratorNonKeyLengthException(f_name,\
+                func_nonkey_argcount, type_nonkey_argcount))
+                
+        if func_key_argcount != type_key_argcount:
+            __raise(DecoratorKeyLengthException(f_name,\
+                func_key_argcount, type_key_argcount))
+
+        # Static check #1:
+        #
+        # kwarg default value types.
+        if func_defaults:
+            tmp_kw_zip = zip(func_key_args, func_defaults)
+            for tmp_kw_name, tmp_kw_default in tmp_kw_zip:
+                if not tmp_kw_name in type_kwargs:
+                    __raise(DecoratorKeyUnspecifiedException(f_name,\
+                         tmp_kw_name))
+                
+                tmp_kw_type = type_kwargs[tmp_kw_name]
+                __check_type(f_name, tmp_kw_name, tmp_kw_type, tmp_kw_default)
+            
+            del tmp_kw_type
+            del tmp_kw_name
+            del tmp_kw_default
+            del tmp_kw_zip
 
         def __wrapper_func(*call_args, **call_kwargs):
-            arg_count = (len(call_args)-start) + len(call_kwargs)
-
-            # Check minimum argument count.
-            if arg_count < func_nonkey_argcount:
-                __raise(ArgumentLengthException(f_name, func_nonkey_argcount,\
-                     arg_count))
-
-            type_nonkey_argcount = len(type_args)
-            type_key_argcount = len(type_kwargs.keys())
             call_nonkey_argcount = len(call_args)
-            
-            type_arg_count = type_nonkey_argcount + type_key_argcount
+            call_key_argcount = len(call_kwargs)
+            call_nonkey_args = []
 
-            # Check argument count of decorator.
-            if type_arg_count != func_argcount:
-                __raise(DecoratorLengthException(f_name, func_argcount,
-                     type_arg_count))
+            if is_method:
+                call_nonkey_args = call_args[1:]
+            else:
+                call_nonkey_args = call_args[:]
 
-            if type_nonkey_argcount != func_nonkey_argcount:
-                __raise(DecoratorNonKeyLengthException(f_name,
-                    func_nonkey_argcount, type_nonkey_argcount))
-
-            # Check non-keyword arguments
+            # Dynamic check #1:
+            #
+            #
+            # Non-keyword argument types.
             if type_nonkey_argcount > 0:
-                for i in range(start, type_nonkey_argcount+start):
-                    arg_name = func.func_code.co_varnames[i]
-                    __check_type(f_name, arg_name, type_args[i-start], call_args[i])
-                
+                tmp_zip = zip(call_nonkey_args, type_args,\
+                     func_nonkey_args)
+                for tmp_call_value, tmp_type, tmp_arg_name in tmp_zip:
+                    __check_type(f_name, tmp_arg_name, tmp_type, tmp_call_value)
 
-            # Check keyword arguments passed as non-keyword arguments
-            for i in range(type_nonkey_argcount+start, call_nonkey_argcount):
-                # Try keyword arguments passed as non-keyword arguments first
-                #kwarg_pos_nonkey = i + (func_nonkey_argcount-start)
-                arg_name = func.func_code.co_varnames[i]
-                arg_value = call_args[i]
-                
-                if not arg_name in type_kwargs:
-                    __raise(DecoratorKeyUnspecifiedException(f_name, arg_name))
-                    continue
 
-                typ = type_kwargs[arg_name]
-                __check_type(f_name, arg_name, typ, arg_value)
+            # Dynamic check #2:
+            #
+            # Keyword arguments passed as non-keyword arguments.
+            if type_nonkey_argcount < call_nonkey_argcount:
+                tmp_kwargs_as_args = zip(call_nonkey_args[type_nonkey_argcount:],\
+                    func_key_args.keys())
 
-                if argvalidate_warn_kwarg_as_arg:
-                    warnings.warn(KWArgAsArgWarning(f_name, arg_name))
+                for tmp_call_value, tmp_kwarg_name in tmp_kwargs_as_args:
+                    tmp_type = type_kwargs[tmp_kwarg_name]
 
-            # Check keyword arguments
-            for arg_name in call_kwargs.keys():
-                if not arg_name in type_kwargs:
-                        __raise(DecoratorKeyUnspecifiedException(f_name,\
-                             arg_name))
+                    if argvalidate_warn_kwarg_as_arg:
+                        warnings.warn(KWArgAsArgWarning(f_name, tmp_kwarg_name))
+
+                    __check_type(f_name, tmp_kwarg_name, tmp_type,\
+                         tmp_call_value)
+
+            # Dynamic check #3:
+            #
+            # Keyword argument types.
+            if call_key_argcount > 0:
+                for tmp_kwarg_name in call_kwargs:
+                    if tmp_kwarg_name not in type_kwargs:
                         continue
 
-                typ = type_kwargs[arg_name]
-                arg_value = call_kwargs[arg_name]
-                __check_type(f_name, arg_name, typ, arg_value)
-
+                    tmp_call_value = call_kwargs[tmp_kwarg_name]
+                    tmp_type = type_kwargs[tmp_kwarg_name]
+                    __check_type(f_name, tmp_kwarg_name, tmp_type,\
+                         tmp_call_value)
+            
             return func(*call_args, **call_kwargs)
 
-        __wrapper_func.func_name = func.func_name
-        __wrapper_func._args_stacked_func = func
+        
+        __wrapper_func.func_name = func.__name__
+        __wrapper_func.__doc__ = func.__doc__
+        __wrapper_func.__dict__.update(func.__dict__)
+
+        __wrapper_func.argvalidate_accepts_stacked_func = func
         return __wrapper_func
     
     return validate
 
-def func_args(*type_args, **type_kwargs):
+def accepts(*type_args, **type_kwargs):
     """
-    Decorator used for checking arguments passed to a function.
+    Decorator used for checking arguments passed to a function or method.
+
+    :param start: method/function-detection override. The number of arguments
+                  defined with start are ignored in all checks.
 
     :param type_args: type definitions of non-keyword arguments.
     :param type_kwargs: type definitions of keyword arguments.
-    
-    
-    :raises ArgumentLengthException: Raised if the number of arguments passed
-        to the function does not match the number of arguments the function
-        accepts.
-        
-    :raises DecoratorLengthException: Raised if the number of arguments 
-        specified in the decorator does not match the number of arguments the 
-        function accepts.
 
     :raises DecoratorNonKeyLengthException: Raised if the number of non-keyword
         arguments specified in the decorator does not match the number of
@@ -431,57 +429,21 @@ def func_args(*type_args, **type_kwargs):
     :raises ArgumentTypeException: Raised if an argument type passed to the
         function does not match the type specified in the decorator.
 
-
-    Example::
-
-      @func_args(int, int, z_is_str_or_int=one_of(int, str))
-      def my_func(x_is_int, y_is_str, z_is_str_or_int=5):
-         pass
-         
-    """
-    return __check_args(type_args, type_kwargs, start=0)
-
-def method_args(*type_args, **type_kwargs):
-    """
-    Decorator used for checking arguments passed to a method.
-
-    :param type_args: type definitions of non-keyword arguments.
-    :param type_kwargs: type definitions of keyword arguments.
-
-    :raises ArgumentLengthException: Raised if the number of arguments passed
-        to the method does not match the number of arguments the method
-        accepts.
-
-    :raises DecoratorLengthException: Raised if the number of arguments
-        specified in the decorator does not match the number of arguments the
-        method accepts.
-
-    :raises DecoratorNonKeyLengthException: Raised if the number of non-keyword
-        arguments specified in the decorator does not match the number of
-        non-keyword arguments the method accepts.
-
-    :raises DecoratorKeyLengthException: Raised if the number of keyword
-        arguments specified in the decorator does not match the number of
-        non-keyword arguments the method accepts.
-
-    :raises DecoratorKeyUnspecifiedException: Raised if a keyword argument's
-        type has not been specified in the decorator.
-
-    :raises ArgumentTypeException: Raised if an argument type passed to the
-        function does not match the type specified in the decorator.
-
     Example::
 
         class MyClass:
-            @method_args(int, str, z_is_str_or_int=one_of(int, str))
-            def my_method(self, x_is_int, y_is_str, z_is_str_or_int=5):
-                pass
+            @accepts(int, str)
+            def my_method(self, x_is_int, y_is_str):
+                [...]
+
+        @accepts(int, str)
+        def my_function(x_is_int, y_is_str):
+            [....]
 
-    Note that the *self* parameter is automatically ignored in the check.
     """
-    return __check_args(type_args, type_kwargs, start=1)
+    return __check_args(type_args, type_kwargs, start=-1)
 
-def return_value(expected_type):
+def returns(expected_type):
     """
     Decorator used for checking the return value of a function or method.
 
@@ -498,23 +460,62 @@ def return_value(expected_type):
             
     """
     def validate(f):
-        return_value_func = getattr(f, '_return_value_stacked_func', None)
-        if return_value_func:
-            raise DecoratorStackingException(return_value_func.func_name,\
-                'return_value')
 
-        func = getattr(f, '_args_stacked_func', f)
+        returns_func = getattr(f, 'argvalidate_returns_stacked_func', None)
+        if returns_func:
+            raise DecoratorStackingException(returns_func.func_name,'returns')
+
+        func = getattr(f, 'argvalidate_accepts_stacked_func', f)
 
         def __wrapper_func(*args, **kwargs):
             result = func(*args, **kwargs)
             __check_return_value(func.func_name, expected_type, result)
             return result
-        
-        __wrapper_func.func_name = func.func_name
-        __wrapper_func._return_value_stacked_func = func
+
+        __wrapper_func.func_name = func.__name__
+        __wrapper_func.__doc__ = func.__doc__
+        __wrapper_func.__dict__.update(func.__dict__)
+            
+        __wrapper_func.argvalidate_returns_stacked_func = func
         return __wrapper_func
+    
     return validate
 
+# Wrappers for old decorators
+def return_value(expected_type):
+    """
+    Wrapper for backwards-compatibility.
+
+    :deprecated: This decorator has been replaced with :func:`returns`.
+
+    """
+    warnings.warn(DeprecationWarning('The return_value decorator has been '\
+        'deprecated. Please use the returns decorator instead.'))
+    return returns(expected_type)
+
+
+def method_args(*type_args, **type_kwargs):
+    """
+    Wrapper for backwards-compatibility.
+
+    :deprecated: This decorator has been replaced with :func:`accepts`.
+
+    """
+    warnings.warn(DeprecationWarning('The method_args decorator has been '\
+        'deprecated. Please use the accepts decorator instead.'))
+    return __check_args(type_args, type_kwargs, start=1)
+
+def func_args(*type_args, **type_kwargs):
+    """
+    Wrapper for backwards-compatibility.
+
+    :deprecated: This decorator has been replaced with :func:`accepts`.
+    """
+    warnings.warn(DeprecationWarning('The func_args decorator has been '\
+        'deprecated. Please use the accepts decorator instead.'))
+    return __check_args(type_args, type_kwargs, start=0)
+
+
 class __OneOfTuple(tuple):
     def __repr__(self):
         return 'one of %r' % (tuple.__repr__(self))
@@ -557,6 +558,7 @@ def warns_kwarg_as_arg():
     This behaviour can be controlled via the environment variable
     :envvar:`ARGVALIDATE_WARN_KWARG_AS_ARG`.
     """
+    return argvalidate_kwarg_as_arg
 
 # Used for readbility, using None alone would be sufficient
 type_any = None
