@@ -11,7 +11,8 @@
 # All configuration values have a default value; values that are commented out
 # serve to show the default value.
 
-import sys, os
+import sys
+import os
 
 file_abspath = os.path.dirname(os.path.abspath(__file__))
 parent_path = os.path.abspath(os.path.join(file_abspath, '..'))
