@@ -17,16 +17,11 @@
 from argvalidate_tests import ArgvalidateTestSuite
 
 try:
-    from sptest import TestMain
-
-    result = TestMain(ArgvalidateTestSuite).run()
+    import nose
+    
+    nose.run(suite=ArgvalidateTestSuite)
 except ImportError:
-    try:
-        import nose
-        
-        nose.run(suite=ArgvalidateTestSuite)
-    except ImportError:
-        import unittest
-        unittest.TextTestRunner(verbosity=2).run(ArgvalidateTestSuite)
+    import unittest
+    unittest.TextTestRunner(verbosity=2).run(ArgvalidateTestSuite)
 
 
